from typing import Optional
from .event import MusicEvent, VoiceEvent
from ..voice import Voice

class ProgramChangeEvent ( VoiceEvent ):
    def __init__ ( self, timestamp : int, voice : Voice, program : int ):
        super().__init__( timestamp, voice )

        self.program : int = program
        
    def __str__ ( self ) -> str:
        return f"[pc=({self.program})]"

class ControlChangeEvent ( VoiceEvent ):
    def __init__ ( self, timestamp : int, voice : Voice, control : int, value : int ):
        super().__init__( timestamp, voice )

        self.control : int = control
        self.value : int = value

    def __str__ ( self ) -> str:
        return f"[cc=({self.control}, {self.value})]"

class ContextChangeEvent( VoiceEvent ):
    def __init__ ( self, timestamp : int, property : str, value, voice : Voice, staff : Optional[int] = 0 ):
        super().__init__( timestamp, voice, staff )

        self.property : str = property
        self.value = value

    def __str__ ( self ) -> str:
        return f"[{self.property}={self.value}]"



from .transformer import Transformer
from typing import List, Union

class TeeTransformer(Transformer):
    def __init__ ( self, sinks : Union[Transformer, List[Transformer]] = [] ):
        if isinstance( sinks, Transformer ):
            self.sinks : List[Transformer] = [ sinks ]
        else:
            self.sinks : List[Transformer] = sinks


    def transform ( self ):
        while True:
            done, event = yield

            if done: break

            for sink in self.sinks:
                sink.add_input( event )
            
            self.add_output( event )
        
        for sink in self.sinks:
            sink.end_input()
        
        self.end_output()





from musikla.parser.printer import CodePrinter
from typing import Tuple
from .music_node import MusicNode
from musikla.core.events import RestEvent

class RestNode( MusicNode ):
    def __init__ ( self, value = None, visible = False, position : Tuple[int, int, int] = None ):
        super().__init__( position )
        
        self.value = value
        self.visible = visible

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( RestEvent( 0, 0, self.value, visible = self.visible ).to_string( ( 'r', 'r' ) ) )

    def get_events ( self, context ):
        rest = RestEvent(
            timestamp = context.cursor,
            duration = context.get_duration( self.value ),
            value = context.get_value( self.value ),
            voice = context.voice,
            visible = self.visible,
            staff = 0
        )

        context.cursor += rest.duration

        yield rest



import sys
import time
from .sequencers import Sequencer, SequencerFactory, ArgumentParser, ArgumentParserError
from musikla.core import Context
from musikla.core.events.event import MusicEvent
from configparser import ConfigParser
from argparse import Namespace
from typing import List, Any, Optional, Union, Iterable
from typing_extensions import Protocol

def get_milliseconds () -> int:
    return int( round( time.time() * 1000 ) )

class PlayerLike(Protocol):
    def get_time( self ) -> int: ...
    
    def play_more ( self, events : Iterable[MusicEvent] ): ...

    def print_error ( self, err : BaseException ): ...

class Player():
    def __init__ ( self, sequencers : List[Sequencer] = [], events = [] ):
        self.events = events
        self.sequencers : List[Sequencer] = sequencers
        self.started : bool = False
        self.start_time : Optional[int] = None
        self.print_events = False
        self.custom_error_printer : Any = None
        
        self.sequencer_factories : List[SequencerFactory] = []

    def print_error ( self, err : BaseException ):
        if self.custom_error_printer is not None:
            self.custom_error_printer( err )

    def add_sequencer_factory ( self, factory : Any, context : Context, config : ConfigParser ):
        self.sequencer_factories.append( factory( context, config ) )

    def make_sequencer_from_format ( self, format : str, uri : str, args : List[str] = [] ) -> Sequencer:
        factory : Optional[SequencerFactory] = None

        for f in self.sequencer_factories:
            if f.name == format:
                factory = f
                break
        
        if factory is None:
            raise Exception( f"Could not create a sequencer for format { format }" )

        if factory.argparser is not None:
            arguments = factory.argparser.parse_args( args )
        else:
            arguments = args

        sequencer = factory.from_str( uri, arguments )

        if sequencer is None:
            raise Exception( f"Could not create a sequencer format { format } for { uri }" )

        return sequencer

    def make_sequencer_from_uri ( self, uri : str, args : List[str] = [] ) -> Sequencer:
        default : Optional[SequencerFactory] = None
        
        sequencer : Optional[Sequencer] = None

        args_errors = []

        for factory in self.sequencer_factories:
            if factory.default:
                if default is not None:
                    print( f"WARNING: Replacing default sequencer { default.__class__.__name__ } with { factory.__class__.__name__ }" )
                    
                default = factory
            else:
                try:
                    if factory.argparser is not None:
                        arguments, _ = factory.argparser.parse_known_args( args )
                    else:
                        arguments = args
                except ArgumentParserError as e:
                    args_errors.append( ( factory.name, str( e ) ) )
                    continue
                
                sequencer = factory.from_str( uri, arguments )

            if sequencer != None: break

        if sequencer is None and default is not None:
            try:
                if default.argparser is not None:
                    arguments, _ = default.argparser.parse_known_args( args )
                else:
                    arguments = args
            except ArgumentParserError as e:
                args_errors.append( ( default.name, str( e ) ) )
            else:
                sequencer = default.from_str( uri, arguments )
        
        if sequencer is None:
            if args_errors:
                sys.stderr.write( f"Could not create a sequencer for { uri }\n" )
                sys.stderr.write( ''.join( [ n + ':\n' + m + '\n' for n, m in args_errors ] ) )
                raise SystemExit( 2 )
            else:
                raise Exception( f"Could not create a sequencer for { uri }" )

        return sequencer

    def add_sequencer ( self, sequencer : Union[Sequencer, str], format = None, args = None ):
        if type( sequencer ) is str:
            if format is None:
                sequencer = self.make_sequencer_from_uri( sequencer, args or [] )
            else:
                sequencer = self.make_sequencer_from_format( format, sequencer, args or [] )

        self.sequencers.append( sequencer )

    @property
    def realtime ( self ) -> bool:
        return any( seq.realtime for seq in self.sequencers )

    def setup ( self ):
        for seq in self.sequencers:
            seq.start()

        self.start_time = get_milliseconds()

        self.started = True

    def get_time ( self ):
        if not self.started:
            return 0
        
        return get_milliseconds() - self.start_time

    def play ( self ):
        if not self.started:
            self.setup()

        for seq in self.sequencers:
            seq.register_events_many( self.events )

    def play_more ( self, events ):
        if not self.started:
            self.setup()

        events = list( events )

        if self.print_events and events:
            print( 'playing', events[ 0 ].timestamp, ' '.join( str( e ) for e in events ), [ e.staff for e in events ] )

        for seq in self.sequencers:
            seq.register_events_many( events )

    def join ( self ):
        if not self.started:
            return

        for seq in self.sequencers:
            seq.join()

    def close ( self ):
        if self.started:
            for seq in self.sequencers:
                seq.close()



from typing import Optional, Tuple
from ..voice import Voice
from .event import DurationEvent
from fractions import Fraction

class RestEvent( DurationEvent ):
    def __init__ ( self, timestamp = 0, duration = 0, value = 0, voice : Voice = None, visible = True, staff : Optional[int] = 0 ):
        super().__init__( timestamp, duration, value, voice, staff )

        self.visible = visible

    def to_string ( self, chars: Tuple[str, str] = ( 'z', 'x' ), append_value : bool = True ):
        rest = chars[ 0 ] if self.visible else chars[ 1 ]

        if append_value and self.value != None and self.value != 1:
            rest += str( Fraction( self.value ) )

        return rest

    def __repr__ ( self ):
        return f'[{self.timestamp}]' + str( self )

    def __str__ ( self ) -> str:
        return self.to_string()



from .transformer import Transformer
from musikla.core import Voice
from musikla.core.events import MusicEvent, NoteEvent, RestEvent, ChordEvent
from typing import List, Optional

class SlidingAverage():
    def __init__ ( self, capacity : int = 0 ):
        self.history : List[float] = []
        self.capacity = capacity
        self.sum : float = 0
        self.count : int = 0

    def __iadd__ ( self, value : float ):
        if self.count == 0:
            self.sum = value
            self.count = 1

            if self.capacity > 0:
                self.history.append( value )
        else:
            #  ( ( self.average / self.count ) + ( value / ( self.count + 1 ) ) )
            self.sum += value

            if self.capacity == 0 or self.count < self.capacity:
                self.count += 1
            else:
                self.sum -= self.history.pop( 0 )

            if self.capacity > 0:
                self.history.append( value )

        return self

    def __int__ ( self ):
        if self.count == 0:
            return 0

        return int( self.sum / self.count )
        
    def __float__ ( self ):
        if self.count == 0:
            return 0

        return float( self.sum / self.count )

class VoiceIdentifierVoice():
    def __init__ ( self, transformer : 'VoiceIdentifierTransformer', parent : Voice, index : int, staff : int = None ):
        self.transformer : VoiceIdentifierTransformer = transformer
        self.parent_name : str = parent.name
        self.parent_staff : Optional[int] = staff
        self.voice : Voice = parent.clone( parent.name + '/' + str( index ) )
        self.index = index
        self.average_pitch : SlidingAverage = SlidingAverage()
        self.auto_create_rests : bool = True
        self.auto_create_end_rests : bool = True
        self.last_event : Optional[MusicEvent] = None

    @property
    def last_timestamp ( self ):
        if self.last_event is None:
            return 0

        return self.last_event.timestamp
    
    @property
    def last_end_timestamp ( self ):
        if not self.last_event:
            return 0

        return self.last_event.end_timestamp

    def is_busy_at ( self, timestamp : int ) -> bool:
        return self.last_end_timestamp > timestamp

    def is_connected_at ( self, timestamp : int ) -> bool:
        return abs( timestamp - self.last_end_timestamp ) <= 3

    def distance ( self, event : NoteEvent ) -> float:
        return abs( int( event ) - int( self.average_pitch ) )

    def create_rest ( self, target : int, visible : bool = True ) -> RestEvent:
        duration = target - self.last_end_timestamp

        value = self.voice.from_duration_absolute( duration )

        return RestEvent(
            timestamp = self.last_end_timestamp,
            visible = visible,
            duration = duration,
            value = value,
            voice = self.voice,
            staff = self.parent_staff
        )

    def append ( self, *events : MusicEvent ):
        for event in events:
            if isinstance( event, NoteEvent ) or isinstance( event, ChordEvent ):
                self.average_pitch += int( event ) if isinstance( event, NoteEvent ) \
                                else  sum( event.pitches ) / len( event.pitches )

                # TODO Allow setting a minimum rest duration. If the empty space between the events is less than
                # said minimum, then no rest is created
                if self.auto_create_rests and self.last_end_timestamp < event.timestamp - 1:
                    self.transformer.add_output( self.create_rest( event.timestamp ) )

            self.transformer.add_output( event.clone( voice = self.voice ) )

            self.last_event = event

class VoiceIdentifierTransformer(Transformer):
    """
    This class receives a flat, ordered stream of musical events and splits them in multiple voices and identifies parallel notes/chords
    """
    def __init__ ( self, auto_create_rests : bool = True, auto_create_end_rests : bool = True ):
        super().__init__()

        self.voices : List[VoiceIdentifierVoice] = []

        self.auto_create_rests : bool = auto_create_rests
        
        self.auto_create_end_rests : bool = auto_create_end_rests
    
    def create_voice_for ( self, event : NoteEvent ) -> VoiceIdentifierVoice:
        index = max( ( voice.index for voice in self.voices if voice.parent_name == event.voice.name ), default = 0 )

        voice = VoiceIdentifierVoice( self, event.voice, index + 1, event.staff )

        voice.auto_create_rests = self.auto_create_rests
        voice.auto_create_end_rests = self.auto_create_end_rests

        self.voices.append( voice )

        return voice

    def find_voice_for ( self, event, auto_create : bool = True ) -> Optional[VoiceIdentifierVoice]:
        best_voice, best_voice_dst = None, None

        for voice in self.voices:
            if voice.parent_name != event.voice.name or voice.parent_staff != event.staff:
                continue

            # If the voice already has a sound playing at this timestamp
            if voice.is_busy_at( event.timestamp ):
                continue

            if event.staff is not None:
                return voice

            voice_dst = voice.distance( event )

            if best_voice is None or best_voice_dst > voice_dst:
                best_voice = voice
                best_voice_dst = voice_dst

        if best_voice is None and auto_create:
            best_voice = self.create_voice_for( event )

        return best_voice
    
    def find_voice_for_rest ( self, event, auto_create : bool = True ) -> VoiceIdentifierVoice:
        for voice in self.voices:
            if voice.parent_name != event.voice.name or voice.parent_staff != event.staff:
                continue

            # If the voice already has a sound playing at this timestamp
            if voice.is_connected_at( event.timestamp ):
                return voice

        return self.create_voice_for( event )

    def register_rest ( self, event : RestEvent ):
        voice = self.find_voice_for_rest( event )

        if voice is not None:
            voice.append( event )

    def register_note ( self, event : NoteEvent ):
        voice = self.find_voice_for( event )

        if voice is not None:
            voice.append( event )
    
    def register_chord ( self, event : ChordEvent ):
        voice = self.find_voice_for( event )

        if voice is not None:
            voice.append( event )

    def transform ( self ):
        """
        For now, NoteOn/NoteOff events will be ignored. 
        Use the compose transformer before passing events into this notation builder. 
        This might change in the future.
        """
        while True:
            done, event = yield

            if done: break

            if isinstance( event, NoteEvent ):
                self.register_note( event )
            elif isinstance( event, RestEvent ):
                self.register_rest( event )
            elif isinstance( event, ChordEvent ):
                self.register_chord( event )
            else:
                self.add_output( event )

        if self.auto_create_end_rests and self.voices:
            last_voice_timestamp : int = max( [ voice.last_end_timestamp for voice in self.voices ] )

            for voice in self.voices:
                if voice.last_end_timestamp < last_voice_timestamp:
                    self.add_output( voice.create_rest( last_voice_timestamp, visible = False ) )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="musikla", # Replace with your own username
    version="0.7.3",
    author="Pedro M. Silva",
    author_email="pemiolsi@hotmail.com",
    description="A programming environment for music accompaniements and keyboards",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pedromsilvapt/miei-dissertation",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    scripts=['bin/musikla'],
    python_requires='>=3.7',
    package_data={ 'musikla': ['parser/grammar.peg', 'parser/grammar.lark', 'libraries/keyboard/library.mkl', 'libraries/std.mkl'] },
    install_requires=[
        'typeguard',
        'pynput',
        'mido',
        'python-rtmidi',
        'lark-parser',
        'colorama',
        'typing_extensions',
        'prompt_toolkit',
        'pygments',
        'pyFluidSynth-musikla'
    ],
    entry_points={
        "console_scripts": [
            "musikla=musikla.__main__:main",
        ]
    },
)



from typing import Tuple
from .. import MusicNode
from musikla.core import Music

class BlockContextModifierNode( MusicNode ):
    def __init__ ( self, body, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.body = body
    
    def modify ( self, context ):
        pass

    def restore ( self, context ):
        pass

    def get_events ( self, context ):
        block_context = context.fork()

        events = self.modify( block_context )

        if isinstance( events, Music ):
            for event in events.expand( context ): yield event

        try:
            if self.body != None:
                events = self.body.eval( block_context )

                if isinstance( events, Music ):
                    for event in events.expand( block_context ): yield event

            events = self.restore( block_context )

            if isinstance( events, Music ):
                for event in events.expand( context ): yield event
        finally:
            context.join( block_context )



from typing import TypeVar, Generic, Iterable, Iterator, Optional, Generator, cast

T = TypeVar( 'T' )

class IterCursor( Generic[T] ):
    def __init__ ( self, iterable : Iterable[T] ):
        self.iterator : Iterator[T] = iter( iterable )
        self.ended : bool = False
        self.current : Optional[T] = None
        
    def move_next ( self ):
        try:
            self.current = next( self.iterator )
            
            return True
        except StopIteration:
            self.current = None

            self.ended = True

            return False

    def close ( self ):
        if callable( getattr( self.iterator, 'close', None ) ):
            cast( Generator, self.iterator ).close()


def merge_sorted ( items, order = lambda x: x ):
    items = map( lambda en: IterCursor( en ), items )
    items = filter( lambda en: en.move_next(), items )
    items = map( lambda enumerator: ( order( enumerator.current ), enumerator ), items )
    items = list( sorted( items, key = lambda en: en[ 0 ] ) )
    
    try:
        while len( items ) > 0:
            next = items[ 0 ]

            yield next[ 1 ].current

            del items[ 0 ]

            if next[ 1 ].move_next():
                value = order( next[ 1 ].current )

                l = len( items )

                for i in range( l + 1 ):
                    if i == l or value < items[ i ][ 0 ]:
                        items.insert( i, ( value, next[ 1 ] ) )

                        break
                
            else: next[ 1 ].close()
    finally:
        for en in items: en[ 1 ].close()

        items = None

__all__ = []



from typing import Optional
from .note import Note
from fractions import Fraction
from typing import List, Dict
from .scales import chords

class Chord:
    # Static
    @staticmethod
    def from_abbreviature ( root_note : Note, name : str, value : Fraction = Fraction() ):
        intervals = chords[ name ]

        notes = [ root_note.clone().transpose( interval ) for interval in intervals ]
        
        return Chord( notes, root_note.to_string( base_octave = 0, append_value = False ) + name, value )

    @staticmethod
    def from_pitches ( pitches : List[int], name : str = None, value : Fraction = Fraction() ):
        notes = [ Note.from_pitch( pitch, value ) for pitch in pitches ]
        
        return Chord( notes, name, value )

    def __init__ ( self, notes : List[Note] = [], name : str = None, value : Fraction = Fraction() ):
        self.notes : List[Note] = notes
        self.name : Optional[str] = name
        self.value : Fraction = value

    def to_pitches ( self ) -> List[int]:
        return [ n.to_pitch() for n in self.notes ]
    
    def timeless ( self ):
        if self.value != None and self.value != 1:
            note = self.clone()
            note.value = None
            return note
        else:
            return self

    def clone ( self ) -> 'Chord':
        return Chord( self.notes, self.name, self.value )

    def __eq__ ( self, other ):
        if other is None:
            return False

        return int( self ) == int( other )

    def __hash__ ( self ):
        return hash( str( self ) )

    def to_string ( self, base_octave : int = 4, append_value : bool = True ) -> str:
        chord = self.name if self.name is not None \
            else ''.join( [ n.to_string( base_octave = base_octave, append_value = False ) for n in self.notes ] )

        chord = '[' + chord + ']'

        if append_value and self.value != None and self.value != 1:
            chord += str( Fraction( self.value ) )

        return chord

    def __str__ ( self ):
        return self.to_string()



from typing import List, Optional, Tuple, Dict, Union
from fractions import Fraction
from musikla.core.theory import NoteAccidental

class ABCVoice:
    def __init__ ( self ):
        self.name : str = None
        self.label : str = None

    @property
    def name_escaped ( self ) -> str:
        return self.name.replace( "/", "_" )

    def __str__ ( self ):
        if self.label is None:
            return f'V:{ self.name_escaped }'
        else:
            return f'V:{ self.name_escaped } name="{ self.label }"'

class ABCVoiceGroupKind:
    Simple = 0
    Parentheses = 1
    CurlyBraces = 2
    Brackets = 3

class ABCVoiceGroup:
    def __init__ ( self, kind : int = ABCVoiceGroupKind.Simple, members : List[Union[str, 'ABCVoiceGroup']] = [], continuous_staves : bool = False ):
        self.kind : int = kind
        self.members : List[Union[str, ABCVoiceGroup]] = members
        self.continuous_staves : bool = continuous_staves
    
    def __contains__ ( self, voice : str ):
        if voice is not str:
            raise Exception( "ABCVoiceGroup can only contain voices as strings, %s given" % str( type( voice ) ) )

        return any( ( m == voice ) if m is str else ( voice in m ) for m in self.members )

    def append ( self,  voice : Union[str, 'ABCVoiceGroup'] ):
        self.members.append( voice )

    def append_with ( self, voice : Union[str, 'ABCVoiceGroup'], other_voice : str ) -> bool:
        """
        Adds this voice to the same group that also contains `other_voice`.
        Returns true if `other_voice` was found and the `voice` was appended, false otherwise
        """
        for m in self.members:
            if m == other_voice:
                self.append( voice )

                return True
            elif isinstance( m, ABCVoiceGroup ) and m.append_with( voice, other_voice ):
                return True

        return False

    def __str__ ( self ):
        non_empty_members = [ m for m in self.members if m ]

        separator = ' | ' if self.continuous_staves else ' '

        score = separator.join( str( m ) for m in non_empty_members )

        if self.kind == ABCVoiceGroupKind.Parentheses:
            return f"({ score })"
        elif self.kind == ABCVoiceGroupKind.CurlyBraces:
            return f"{{{ score }}}"
        elif self.kind == ABCVoiceGroupKind.Brackets:
            return f"[{ score }]"
        else:
            return score

    def __bool__ ( self ):
        return any( [ bool( m ) for m in self.members ] )

class ABCVoiceScore:
    def __init__ ( self ):
        self.groups : ABCVoiceGroup = ABCVoiceGroup()
        self.verbatim : Optional[str] = None

    def __str__ ( self ):
        has_groups = bool( self.groups )
        has_verbatim = self.verbatim is not None

        if has_groups and has_verbatim:
            return f"{ str( self.groups ) } { self.verbatim }"
        elif has_groups:
            return str( self.groups )
        elif has_verbatim:
            return self.verbatim
        else:
            return ""

    def __bool__ ( self ):
        return bool( self.groups ) or bool( self.verbatim )

class ABCHeader:
    def __init__ ( self ):
        # X:1
        self.reference : int = 1
        # T:Paddy O'Rafferty
        self.title : Optional[str] = None
        # C:Trad.
        self.composer : Optional[str] = None
        # M:6/8
        self.meter : Optional[Tuple[int, int]] = None
        # L:1/8
        self.length : Optional[Fraction] = None
        # Q:74
        self.tempo : Optional[int] = None
        # K:D
        self.key : str = 'C' # Must always be the last field of the header
        # V:V1
        self.voices : List[ABCVoice] = []

        self.score : ABCVoiceScore = ABCVoiceScore()

    def __str__ ( self ):
        parts : List[str] = []

        if self.reference is not None:
            parts.append( f"X:{ self.reference }" )

        if self.title is not None:
            parts.append( f"T:{ self.title }" )

        if self.composer is not None:
            parts.append( f"C:{ self.composer }" )

        if self.meter is not None:
            parts.append( f"M:{ self.meter[ 0 ] }/{ self.meter[ 1 ] }" )
        
        if self.length is not None and self.length != 1:
            parts.append( f"L:{ self.length }" )

        if self.tempo is not None:
            parts.append( f"Q:{ self.tempo }" )

        for voice in self.voices:
            parts.append( str( voice ) )

        if self.score:
            parts.append( f"%%score { str( self.score ) }" )

        if self.key is not None:
            parts.append( f"K:{ self.key }" )

        return '\n'.join( parts )

class ABCSymbol:
    def __init__ ( self ):
        self.length : Optional[Fraction] = None

class ABCNote(ABCSymbol):
    def __init__ ( self ):
        super().__init__()

        self.pitch_class : str = None
        self.octave : int = None
        self.accidental : int = NoteAccidental.NONE
        self.tied : bool = False

    def __str__ ( self ):
        note : str = self.pitch_class.lower()

        if self.octave <= 3:
            note = note.upper()

            for _ in range( 2, self.octave - 1, -1 ): note += ","
        else:
            for _ in range( 5, self.octave + 1 ): note += "'"
        
        if self.length != None and self.length != 1:
            note += str( self.length )

        if self.tied:
            note += '-'

        if self.accidental == NoteAccidental.DOUBLESHARP:
            note = '^^' + note
        elif self.accidental == NoteAccidental.SHARP:
            note = '^' + note
        elif self.accidental == NoteAccidental.FLAT:
            note = '_' + note
        elif self.accidental == NoteAccidental.DOUBLEFLAT:
            note = '__' + note

        return note

class ABCChord(ABCSymbol):
    def __init__ ( self ):
        super().__init__()

        self.notes : List[ABCNote] = []
        self.name : Optional[str] = None
        self.tied : bool = False

    def __str__ ( self ):
        notes = ''.join( str( n ) for n in self.notes )

        chord = f"[{ notes }]"

        if self.name is not None:
            chord = f'"{ self.name }"{ chord }'
        
        if self.length != None and self.length != 1:
            chord += str( self.length )

        if self.tied:
            chord += '-'

        return chord
        

class ABCRest(ABCSymbol):
    def __init__ ( self ):
        self.visible : bool = True

    def __str__ ( self ):
        rest = 'z' if self.visible else 'x'

        if self.length != None and self.length != 1:
            rest += str( self.length )

        return rest

class ABCBar:
    def __init__ ( self ):
        self.symbols : List[ABCSymbol] = []
    
    def __str__ ( self ):
        return ' '.join( [ str( symbol ) for symbol in self.symbols ] )

    @property
    def length ( self ) -> Fraction:
        if len( self.symbols ) == 0:
            return Fraction()
        
        return Fraction( sum( [ s.length for s in self.symbols if s.length != None ] ) )

    def __len__ ( self ):
        return len( self.symbols )

class ABCStaff:
    def __init__ ( self ):
        self.bars : List[ABCBar] = []

    def __str__ ( self ):
        return ' | '.join( [ str( bar ) for bar in self.bars if bar ] ) + ' |'

    def __len__ ( self ):
        return sum( len( b ) for b in self.bars )

class ABCBody:
    def __init__ ( self ):
        self.staffs : Dict[str, List[ABCStaff]] = {}

    def __str__ ( self ):
        return '\n'.join( [ f'[V:{voice.replace( "/", "_" )}] ' + str( staff ) for voice, staffs in self.staffs.items() for staff in staffs if staff ] )

# dff cee|def gfe|dff cee|dfe dBA|dff cee|def gfe|faf gfe|1 dfe dBA:|2 dfe dcB|]
# ~A3 B3|gfe fdB|AFA B2c|dfe dcB|~A3 ~B3|efe efg|faf gfe|1 dfe dcB:|2 dfe dBA|]
# fAA eAA|def gfe|fAA eAA|dfe dBA|fAA eAA|def gfe|faf gfe|dfe dBA:|
class ABCFile:
    def __init__ ( self ):
        self.header : ABCHeader = ABCHeader()
        self.body : ABCBody = ABCBody()

    def __str__ ( self ):
        return f"{ self.header }\n{ self.body }\n"



"""
A simple example of a Notepad-like text editor.
"""
import datetime
from asyncio import Future, ensure_future, sleep, create_task
from typing import Any, List, Optional, cast
from musikla.core import Context, Value, Music
from musikla.audio import Player, InteractivePlayer
from musikla.parser import Parser
import re

from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.application import Application
from prompt_toolkit.application.current import get_app
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Float,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.completion import Completer, Completion, FuzzyCompleter
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.lexers import DynamicLexer, PygmentsLexer
from prompt_toolkit.search import start_search
from prompt_toolkit.styles import Style
from prompt_toolkit.document import Document
from prompt_toolkit.widgets import (
    Button,
    Dialog,
    Label,
    MenuContainer,
    MenuItem,
    SearchToolbar,
    TextArea,
)


class ApplicationState:
    """
    Application state.

    For the simplicity, we store this as a global, but better would be to
    instantiate this as an object and pass at around.
    """

    context : Optional[Context] =  None
    player : Optional[Player] = None
    document : Optional[Document] = None
    interactive_players : List[InteractivePlayer] = []
    show_status_bar = True
    current_path = None

MUSIKLA_WORD_RE = re.compile(r"[a-zA-Z0-9_\\]*\$?")

class VariableCompleter(Completer):
    def __init__ ( self, context : Optional[Context] ):
        self.context : Optional[Context] = context

    def get_completions(self, document, complete_event):
        if self.context is None:
            return

        word = document.get_word_before_cursor(pattern = MUSIKLA_WORD_RE)
        
        if len(word) > 0 and word[0] == '$':
            for name, value in self.context.symbols.enumerate( prefix = word[ 1: ] ):
                yield Completion(
                    '$' + name,
                    start_position=-len(word),
                    # style="fg:" + color,
                    # selected_style="fg:white bg:" + color,
                )
        elif len(word) > 0 and word[ 0 ] != '$':
            for name, value in self.context.symbols.enumerate( prefix = word ):
                if callable( value ):
                    yield Completion(
                        name + '()',
                        start_position=-len(word),
                        # style="fg:" + color,
                        # selected_style="fg:white bg:" + color,
                    )

def create_application():
    def get_statusbar_text():
        return "Ctrl-C: Close. Ctrl-E: Execute. Ctrl-S: Stop music. F2: Open Menu. "

    def get_statusbar_right_text():
        return " {}:{}  ".format(
            text_field.document.cursor_position_row + 1,
            text_field.document.cursor_position_col + 1,
        )


    search_toolbar = SearchToolbar()
    text_field = TextArea(
        lexer=PygmentsLexer.from_filename( '.mkl', sync_from_start=False ),
        completer = VariableCompleter(ApplicationState.context),
        scrollbar=True,
        line_numbers=True,
        search_field=search_toolbar,
    )


    class TextInputDialog:
        def __init__(self, title="", label_text="", completer=None):
            self.future = Future()

            def accept_text(buf):
                get_app().layout.focus(ok_button)
                buf.complete_state = None
                return True

            def accept():
                self.future.set_result(self.text_area.text)

            def cancel():
                self.future.set_result(None)

            self.text_area = TextArea(
                completer=completer,
                multiline=False,
                width=D(preferred=40),
                accept_handler=accept_text,
            )

            ok_button = Button(text="OK", handler=accept)
            cancel_button = Button(text="Cancel", handler=cancel)

            self.dialog = Dialog(
                title=title,
                body=HSplit([Label(text=label_text), self.text_area]),
                buttons=[ok_button, cancel_button],
                width=D(preferred=80),
                modal=True,
            )

        def __pt_container__(self):
            return self.dialog


    class MessageDialog:
        def __init__(self, title, text):
            self.future = Future()

            def set_done():
                self.future.set_result(None)

            ok_button = Button(text="OK", handler=(lambda: set_done()))

            self.dialog = Dialog(
                title=title,
                body=HSplit([Label(text=text),]),
                buttons=[ok_button],
                width=D(preferred=80),
                modal=True,
            )

        def __pt_container__(self):
            return self.dialog


    body = HSplit(
        [
            text_field,
            search_toolbar,
            ConditionalContainer(
                content=VSplit(
                    [
                        Window(
                            FormattedTextControl(get_statusbar_text), style="class:status"
                        ),
                        Window(
                            FormattedTextControl(get_statusbar_right_text),
                            style="class:status.right",
                            width=9,
                            align=WindowAlign.RIGHT,
                        ),
                    ],
                    height=1,
                ),
                filter=Condition(lambda: ApplicationState.show_status_bar),
            ),
        ]
    )

    # Global key bindings.
    bindings = KeyBindings()


    @bindings.add("f2")
    def _(event):
        " Focus menu. "
        event.app.layout.focus(root_container.window)


    #
    # Handlers for menu items.
    #


    def do_open_file():
        async def coroutine():
            open_dialog = TextInputDialog(
                title="Open file",
                label_text="Enter the path of a file:",
                completer=PathCompleter(),
            )

            path = await show_dialog_as_float(open_dialog)
            ApplicationState.current_path = path

            if path is not None:
                try:
                    with open(path, "rb") as f:
                        text_field.text = f.read().decode("utf-8", errors="ignore")
                except IOError as e:
                    show_message("Error", "{}".format(e))

        ensure_future(coroutine())


    def do_about():
        show_message("About", "Text editor demo.\nCreated by Jonathan Slenders.")


    def show_message(title, text):
        async def coroutine():
            dialog = MessageDialog(title, text)
            
            await show_dialog_as_float(dialog)

        ensure_future(coroutine())


    async def show_dialog_as_float(dialog):
        " Coroutine. "
        float_ = Float(content=dialog)
        root_container.floats.insert(0, float_)

        app = get_app()

        focused_before = app.layout.current_window
        app.layout.focus(dialog)
        result = await dialog.future
        app.layout.focus(focused_before)

        if float_ in root_container.floats:
            root_container.floats.remove(float_)

        return result

    @bindings.add( 'c-n' )
    def do_new_file(event = None):
        text_field.text = ""

    @bindings.add( 'c-c' )
    def do_exit(event = None):
        get_app().exit()


    def do_time_date():
        text = datetime.datetime.now().isoformat()
        text_field.buffer.insert_text(text)


    def do_go_to():
        async def coroutine():
            dialog = TextInputDialog(title="Go to line", label_text="Line number:")

            line_number = await show_dialog_as_float(dialog)

            try:
                line_number = int(line_number)
            except ValueError:
                show_message("Invalid line number")
            else:
                text_field.buffer.cursor_position = text_field.buffer.document.translate_row_col_to_index(
                    line_number - 1, 0
                )

        ensure_future(coroutine())


    def do_undo():
        text_field.buffer.undo()


    def do_cut():
        data = text_field.buffer.cut_selection()
        get_app().clipboard.set_data(data)


    def do_copy():
        data = text_field.buffer.copy_selection()
        get_app().clipboard.set_data(data)


    def do_delete():
        text_field.buffer.cut_selection()


    def do_find():
        start_search(text_field.control)


    def do_find_next():
        search_state = get_app().current_search_state

        cursor_position = text_field.buffer.get_search_position(
            search_state, include_current_position=False
        )
        text_field.buffer.cursor_position = cursor_position


    def do_paste():
        text_field.buffer.paste_clipboard_data(get_app().clipboard.get_data())


    def do_select_all():
        text_field.buffer.cursor_position = 0
        text_field.buffer.start_selection()
        text_field.buffer.cursor_position = len(text_field.buffer.text)

    @bindings.add("c-e")
    def do_eval (event = None):
        if ApplicationState.context is None or ApplicationState.player is None:
            show_message("Error", "Cannot evaluate because no context is defined")

            return

        code = text_field.buffer.text

        try:
            script = ApplicationState.context.script

            now = ApplicationState.player.get_time()

            ctx = ApplicationState.context.fork( cursor = now )

            val = script.eval( code, context = ctx )

            if val is not None and isinstance( val, Music ):
                pl = InteractivePlayer( lambda: val.expand( ctx ), ApplicationState.player, realtime = True )

                ApplicationState.interactive_players.append( pl )

                create_task( pl.start() )
        except BaseException as e:
            if hasattr(e, 'message'):
                show_message("Eval Error", str(cast(Any, e).message))
            else:
                show_message("Eval Error", str(e))

    @bindings.add("c-s")
    def do_mute_players(event = None):
        for p in ApplicationState.interactive_players:
            create_task( p.stop() )
        
        ApplicationState.interactive_players = []


    def do_status_bar():
        ApplicationState.show_status_bar = not ApplicationState.show_status_bar

    #
    # The menu container.
    #


    root_container = MenuContainer(
        body=body,
        menu_items=[
            MenuItem(
                "File",
                children=[
                    MenuItem("New...", handler=do_new_file),
                    MenuItem("Open...", handler=do_open_file),
                    MenuItem("Save"),
                    MenuItem("Save as..."),
                    MenuItem("-", disabled=True),
                    MenuItem("Exit", handler=do_exit),
                ],
            ),
            MenuItem(
                "Edit",
                children=[
                    MenuItem("Undo", handler=do_undo),
                    MenuItem("Cut", handler=do_cut),
                    MenuItem("Copy", handler=do_copy),
                    MenuItem("Paste", handler=do_paste),
                    MenuItem("Delete", handler=do_delete),
                    MenuItem("-", disabled=True),
                    MenuItem("Find", handler=do_find),
                    MenuItem("Find next", handler=do_find_next),
                    MenuItem("Replace"),
                    MenuItem("Go To", handler=do_go_to),
                    MenuItem("Select All", handler=do_select_all),
                    MenuItem("Time/Date", handler=do_time_date),
                ],
            ),
            MenuItem("View", children=[MenuItem("Status Bar", handler=do_status_bar),]),
            MenuItem("Info", children=[MenuItem("About", handler=do_about),]),
        ],
        floats=[
            Float(
                xcursor=True,
                ycursor=True,
                content=CompletionsMenu(max_height=16, scroll_offset=1),
            ),
        ],
        key_bindings=bindings,
    )


    style = Style.from_dict({"status": "reverse", "shadow": "bg:#440044",})


    layout = Layout(root_container, focused_element=text_field)

    application = Application(
        layout=layout,
        enable_page_navigation_bindings=True,
        style=style,
        mouse_support=True,
        full_screen=True,
    )

    return application, text_field

def load_state ( application, text_field ):
    if ApplicationState.document is not None:
        text_field.document = ApplicationState.document
    
def save_state ( application, text_field ):
    ApplicationState.document = text_field.document

def run():
    application, text_field = create_application()

    load_state(application, text_field)
    application.run()
    save_state(application, text_field)
    application.input.close()

async def run_async ( context : Context, player : Player ):
    if ApplicationState.context is not None:
        raise Exception( "Musikla Eval Prompt is not reentrant" )

    ApplicationState.context = context
    ApplicationState.player = player

    try:
        with patch_stdout():
            application, text_field = create_application()

            with application.input.raw_mode():
                application.input.read_keys()

            load_state(application, text_field)
            await application.run_async()
            save_state(application, text_field)
    finally:
        ApplicationState.context = None
        ApplicationState.player = None

if __name__ == "__main__":
    run()



from musikla.parser.printer import CodePrinter
from typing import Tuple
from .block_context_modifier_node import BlockContextModifierNode
from musikla.core.events import ProgramChangeEvent
from musikla.core import Value, Voice, Context, TemplateMusic

class VoiceBlockModifier( BlockContextModifierNode ):
    def __init__ ( self, body, voice_name : str, position : Tuple[int, int, int] = None ):
        super().__init__( body, position )

        self.voice_name : str = voice_name

    def modify ( self, context ):
        voice : Voice = context.symbols.lookup( self.voice_name )

        Value.expect( voice, Voice, "Voice modifier " + self.voice_name )
        
        context.voice = voice

    def restore ( self, context ):
        pass

    def __eval__ ( self, context : Context ):
        if self.voice_name == '?':
            return TemplateMusic( self.body )
        else:
            return super().__eval__( context )
    
    def to_source ( self, printer : CodePrinter ):
        with printer.block( '(', ')' ):
            printer.add_token( ':' + str( self.voice_name ) + ' ' )

            self.body.to_source( printer )



from musikla.core.events.event import MusicEvent
from musikla.core.voice import Voice
from .transformer import Transformer
from typing import List, Tuple

def find ( arr, predicate, default_value = None ):
    for elem in arr:
        if predicate( elem ):
            return elem

    return default_value

class NotationBuilderTransformer(Transformer):
    def __init__ ( self, only_final : bool = False ):
        super().__init__()

        self.only_final : bool = only_final

    def to_string ( self, events_per_voice : List[Tuple[Voice, int, List[MusicEvent]]] ) -> str:
        raise NotImplementedError()
    
    def transform ( self ):
        events_per_voice : List[Tuple[Voice, int, List[MusicEvent]]] = []

        while True:
            done, event = yield

            if done: break

            name = event.voice.name

            voice_events = find( events_per_voice, lambda pair: pair[ 0 ].name == name and pair[ 1 ] == event.staff )

            if voice_events is None:
                voice_events = ( event.voice, event.staff, [] )

                events_per_voice.append( voice_events )

            voice_events[ 2 ].append( event )

            if not self.only_final:
                self.add_output( self.to_string( events_per_voice ) )

        if self.only_final:
            self.add_output( self.to_string( events_per_voice ) )



from musikla.core.events import MusicEvent, VoiceEvent, ContextChangeEvent, NoteEvent, ChordEvent, RestEvent, BarNotationEvent, StaffNotationEvent
from musikla.core import MusicBuffer, Voice
from .transformer import Transformer
from typing import Optional, Tuple, Dict, cast
from fractions import Fraction

TimeSignature = Tuple[int, int]

class VoiceNotationInfo:
    def __init__ ( self, name : str, time_signature : TimeSignature, bars_per_staff : int, beats_per_bar : int ):
        self.name : str = name
        self.time_signature = time_signature
        self.bars_per_staff = bars_per_staff
        self.beats_per_bar = beats_per_bar
        self._beat_value : Optional[Fraction] = None

        self.beats_count : float = 0
        self.bars_count : int = 0

    @property
    def is_staff_full ( self ) -> bool:
        return self.bars_count >= self.bars_per_staff - 1 and self.is_bar_full

    @property
    def is_bar_full ( self ) -> bool:
        return self.beats_count >= self.beats_per_bar

    @property
    def beat_value ( self ) -> Fraction:
        if self._beat_value is None:
            ( u, l ) = self.time_signature

            if u >= 6 and u % 3 == 0:
                self._beat_value = Fraction( 3, 2 * l )
            else:
                self._beat_value = Fraction( 1, l )

        return self._beat_value

    def voice_value_to_beats ( self, voice : Voice, value ) -> float:
        return value / self.beat_value
    
    def beats_to_voice_value ( self, beats, voice : Voice ) -> float:
        return beats * self.beat_value

    def split_value ( self, voice : Voice, value ):
        if self.is_bar_full:
            yield 0, 0, False

        event_beats = self.voice_value_to_beats( voice, value )

        leftovers = self.beats_per_bar - self.beats_count


        if event_beats <= leftovers:
            yield value, voice.get_duration_absolute( value ), True

            self.beats_count += event_beats
        else:
            while event_beats > leftovers:
                sub_value = self.beats_to_voice_value( leftovers, voice )
                
                sub_duration = voice.get_duration_absolute( sub_value )

                yield sub_value, sub_duration, False
                
                self.beats_count += leftovers

                event_beats -= leftovers

                leftovers = self.beats_per_bar - self.beats_count

            sub_value = self.beats_to_voice_value( event_beats, voice )
                
            sub_duration = voice.get_duration_absolute( sub_value )

            yield sub_value, sub_duration, True

            self.beats_count += event_beats

    def new_staff ( self ):
        self.bars_count = 0
        self.beats_count = 0

    def new_bar ( self ):
        self.bars_count += 1
        self.beats_count = 0

    def transform ( self, event : MusicEvent ):
        if isinstance( event, RestEvent ) or isinstance( event, ChordEvent ) or isinstance( event, NoteEvent ):
            # Analyze how much duration this measure has left, and if necessary, split up the rest event into two
            # And then pass them along through the `self.add_output` method
            timestamp = event.timestamp

            for sub_value, sub_duration, is_last in self.split_value( event.voice, event.value ):
                if self.is_staff_full:
                    self.new_staff()

                    yield StaffNotationEvent( timestamp, event.voice, event.staff )
                elif self.is_bar_full:
                    self.new_bar()

                    yield BarNotationEvent( timestamp, event.voice, event.staff )

                if sub_value > 0:
                    if isinstance( event, RestEvent ):
                        yield event.clone( timestamp = timestamp, value = sub_value, duration = sub_duration )
                    else:
                        # Keep track of the length of the notes that are played and when a new measure needs to be added
                        # Also analyze how much duration this measure has left, and if necessary, split up the note event into two
                        # and mark the first part of it as `tied`, so that the it is known the next note is to be played continuously
                        yield event.clone( timestamp = timestamp, value = sub_value, duration = sub_duration, tied = not is_last )

                    timestamp += sub_duration
        else:
            yield event


class AnnotateTransformer(Transformer):
    """
    This transformer expects composed notes. Also this transformer expects that note events are either contiguous,
    or separated by rest events for each voice. If that is not the case, yout can use the voice identifier transformer before this one.
    """
    def __init__ ( self, time_signature : TimeSignature = None ):
        super().__init__()

        self.time_signature : Optional[TimeSignature] = time_signature
        self.voices : Dict[str, VoiceNotationInfo] = {}
        self.buffered_events : MusicBuffer = MusicBuffer()
        
    @property
    def bars_per_staff ( self ) -> int:
        # TODO Temporary, should be configurable and with a sensible default
        return 2
        return 16 // self.time_signature[ 0 ]

    @property
    def beats_per_bar ( self ) -> int:
        return self.time_signature[ 0 ]

    def get_voice_for ( self, event : VoiceEvent ) -> VoiceNotationInfo:
        name = event.voice.name
        
        if name not in self.voices:
            self.voices[ name ] = VoiceNotationInfo( name, cast( TimeSignature, self.time_signature ), self.bars_per_staff, self.beats_per_bar )
        
        return self.voices[ name ]

    def transform ( self ):
        while True:
            done, event = yield

            if done: break

            if self.time_signature is None and isinstance( event, ContextChangeEvent ) and event.property == 'timeSignature':
                self.time_signature = event.value

            if self.time_signature is None and isinstance( event, VoiceEvent ):
                self.time_signature = event.voice.time_signature
                
            for buffered_event in self.buffered_events.collect( event.timestamp ):
                self.add_output( buffered_event )

            if isinstance( event, VoiceEvent ):
                voice_info = self.get_voice_for( event )

                # Transform this event and possibly create any others that might be needed
                # Note that instead of emitting them forward immediately, we add them to the buffer
                # This is because some of them might have timestamps in the future, and all the events must be propertly sorted
                # This means that we have to wait for the next inputs before being sure we can emit the new ones
                for ev in voice_info.transform( event ):
                    self.buffered_events.append( ev )
            else:
                self.add_output( event )

        for buffered_event in self.buffered_events.collect():
            self.add_output( buffered_event )







from musikla.core.events.transformers.transformer import Transformer
from musikla.core.events.transformers.balance_notes import BalanceNotesTransformer
from musikla.core.events.transformers.compose_notes import ComposeNotesTransformer
from musikla.audio.player import PlayerLike
from musikla.audio.interactive_player import InteractivePlayer
from musikla.core.context import Context
from musikla.core.music import Music, MusicBuffer
from musikla.core.events import MusicEvent
from typing import Dict, List, Optional, cast
from .keyboard import Keyboard

class KeyboardBuffer:
    @staticmethod
    def save_all ( context : Context, file : str, buffers ):
        from musikla.parser.printer import CodePrinter
        from musikla.libraries.music import function_to_mkl
        import musikla.parser.abstract_syntax_tree as ast
        import musikla.parser.abstract_syntax_tree.statements as ast_st
        import musikla.parser.abstract_syntax_tree.expressions as ast_ex

        body = ast_st.StatementsListNode( [] )

        body.statements.append( ast_st.VariableDeclarationStatementNode( 'buffers', ast_ex.ObjectLiteralNode( [] ) ) )

        for key, buffer in buffers.items():
            set_fn = ast_ex.PropertyAccessorNode( ast_ex.VariableExpressionNode( 'buffers' ), ast_ex.StringLiteralNode( 'set' ), True )

            key_p = ast_ex.NumberLiteralNode( key )
            music_p = function_to_mkl( context, buffer.to_music(), ast = True ) \
                if buffer else \
                ast_ex.NoneLiteralNode()

            body.statements.append( ast_ex.FunctionExpressionNode( set_fn, [ key_p, music_p ] ) )

        file_str = CodePrinter().print( body )

        with open( file, 'w+' ) as f:
            f.write( file_str )
    
    @staticmethod
    def load_all ( context : Context, file : str, buffers = None ):
        from musikla.parser.abstract_syntax_tree.expressions.object_literal_node import Object

        buffers = Object() if buffers is None else buffers

        forked_context = context.fork( symbols = context.symbols.fork() )

        context.script.import_module( forked_context, file, save_cache = False )

        saved_buffers = forked_context.symbols.lookup( "buffers", recursive = False )
        
        for key, music in saved_buffers.items():
            bf = buffers.get( key, default = None )

            if bf is None:
                bf = KeyboardBuffer( context, start = False )
            
            if music is not None:
                bf.from_music( music )

            buffers.set( key, bf )
        
        return buffers

    def __init__ ( self, context : Context, keyboards : List[Keyboard] = None, start : bool = True ):
        from .library import KeyboardLibrary
        lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

        self.context = context
        self.keyboards : List[Keyboard] = []
        self.music_buffer : Optional[MusicBuffer] = None
        self.start_time : Optional[int] = None
        self.collected_events : List[MusicEvent] = []
        self.player : PlayerLike = lib.player
        
        self.loop : bool = False
        self.mode : str = 'music'

        if keyboards is None:
            self.keyboards = lib.keyboards
        else:
            self.keyboards = keyboards

        if start: self.start()
    
    @property
    def duration ( self ) -> int:
        if self.collected_events:
            return self.collected_events[ -1 ].end_timestamp
        else:
            return 0

    @property
    def recording ( self ) -> bool:
        return self.music_buffer is not None

    def on_new_player ( self, keyboard : Keyboard, player : InteractivePlayer ):
        if self.music_buffer is not None:
            if player.buffers is not None:
                player.buffers.append( self.music_buffer )
            else:
                player.buffers = [ self.music_buffer ]

    def start ( self ):
        if self.music_buffer is None:
            self.start_time = self.player.get_time()
            
            self.music_buffer = MusicBuffer()

            for kb in self.keyboards: kb.add_new_player_observer( self.on_new_player )

    def stop ( self, discard : bool = False ):
        if self.music_buffer is not None:
            if not discard:

                collected : List[MusicEvent] = list( self.music_buffer.collect() )

                if collected:
                    # We consider the start time as being the first event's timestamp
                    st : int = collected[ 0 ].timestamp

                    collected = [ ev.clone( timestamp = ev.timestamp - st + self.duration ) for ev in collected ]

                    self.collected_events.extend( collected )

            self.music_buffer = None

            for kb in self.keyboards: kb.remove_new_player_observer( self.on_new_player )

            self.start_time = None
    
    def clear ( self ):
        self.stop( discard = True )

        self.collected_events.clear()

    def toggle ( self, clear : bool = True ):
        if self.recording:
            self.stop()
        else:
            if clear: self.clear()

            self.start()

    def __len__ ( self ):
        return len( self.collected_events )

    def __bool__ ( self ):
        return bool( self.collected_events )

    @property
    def duration_live ( self ):
        # Note that this method might return lower values than previous calls
        
        # When we are recording and we have at least on 
        if self.music_buffer is not None and self.music_buffer.buffer:
            st : int = self.music_buffer.buffer[ 0 ].timestamp

            return self.player.get_time() - st + self.duration
        else:
            return self.duration

    def to_events ( self ):
        events = self.collected_events
        events = BalanceNotesTransformer.iter( events )
        events = ComposeNotesTransformer.iter( events )

        return events

    def to_events_live ( self, get_time = None, now : bool = False ):
        if self.music_buffer is not None:
            buffered : List[MusicEvent] = self.music_buffer.buffer

            if buffered:
                # We consider the start time as being the first event's timestamp
                st : int = buffered[ 0 ].timestamp

                buffered = [ ev.clone( timestamp = ev.timestamp - st + self.duration ) for ev in buffered ]

                if now and get_time is None:
                    st : int = buffered[ 0 ].timestamp

                    get_time = lambda: self.duration_live


            events = self.collected_events + buffered
        
            events = BalanceNotesTransformer.iter( events, get_time = get_time )
            events = ComposeNotesTransformer.iter( events )

            return events
        else:
            return self.to_events()

    def to_music ( self ):
        return Music( self.to_events() ).shared()

    def to_music_live ( self ):
        return Music( self.to_events_live() ).shared()

    def from_music ( self, music : Music ):
        self.clear()

        self.collected_events = list( music )